"""
The Mid-latitude All-sky-imaging Network for Geophysical Observations (MANGO) employs a combination of two powerful optical techniques used to observe the dynamics of Earth's upper atmosphere: wide-field imaging and high-resolution spectral interferometry. Both techniques observe the naturally occurring airglow emissions produced in the upper atmosphere at 630.0- and 557.7-nm wavelengths. Instruments are deployed to sites across the continental United States, providing the capability to make measurements spanning mid to sub-auroral latitudes. The current instrument suite in MANGO has six all-sky imagers (ASIs) observing the 630.0-nm emission (integrated between ~200 and 400 km altitude), six ASIs observing the 557.7-nm emission (integrated between ~90 and 100 km altitude), and four Fabry-Perot interferometers measuring neutral winds and temperature at these wavelengths. The deployment of additional imagers is planned. The network makes unprecedented observations of the nighttime thermosphere-ionosphere dynamics with the expanded field-of-view provided by the distributed network of instruments. 

Instrument paper: https://doi.org/10.1029/2023JA031589
"""
from datetime import datetime, timedelta, timezone
import re
import warnings
import pathlib
import copy
import os
from typing import Tuple, Iterable, List, Union

import requests
import matplotlib.colors
import pandas as pd
import h5py
import numpy as np

import asilib
import asilib.map
import asilib.utils as utils
import asilib.download as download

base_url = 'https://data.mangonetwork.org/data/transport/mango/archive/'
local_base_dir = asilib.config['ASI_DATA_DIR'] / 'mango'


def mango(
    location_code: str,
    channel: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    alt: int = 110,
    acknowledge: bool = True,
    redownload: bool = False,
    missing_ok: bool = True,
    imager=asilib.Imager,
) -> asilib.Imager:
    """
    Create an Imager instance with the MANGO ASI images and skymaps.

    The imaging data are obtained through the MANGO network and operated by SRI with 
    support from US National Science Foundation award AGS-1933013. Please reach out
    to the PI Asti Bhatt (asti.bhatt@sri.com) before using the data for publication.
    Cite the `instrument paper <https://doi.org/10.1029/2023JA031589>`_ for the 
    description of the MANGO network.

    For more information see: https://www.mangonetwork.org/

    Parameters
    ----------
    location_code: str
        The three-letter ASI location code.
    channel: str
        The color channel. Could be "redline", "greenline", "r", or "g". Case insensitive. 
    time: str or datetime.datetime
        A time to look for the ASI data at. Either time or time_range
        must be specified (not both or neither).
    time_range: list of str or datetime.datetime
        A length 2 list of string-formatted times or datetimes to bracket
        the ASI data time interval.
    alt: int
        The reference skymap altitude, in kilometers.
    acknowledge: bool
        If True, prints the acknowledgment statement for MANGO. 
    redownload: bool
        If True, will download the data from the internet, regardless of
        wether or not the data exists locally (useful if the data becomes
        corrupted).
    imager: asilib.Imager
        Controls what Imager instance to return, asilib.Imager by default. This
        parameter is useful if you need to subclass asilib.Imager.

    Examples
    --------
    # Animate a MANGO map with a SAR arc, as well as the SymH index.
    >>> # You will need to install cdasws to run this example (python -m pip install cdasws)
    >>> from datetime import datetime, timedelta, timezone
    >>> 
    >>> import cdasws
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib.dates
    >>> import pandas as pd
    >>> import asilib.asi
    >>> 
    >>> time_range=(datetime(2021, 11, 4, 1, 0), datetime(2021, 11, 4, 12, 24))
    >>> location_code='CFS'
    >>> asi = asilib.asi.mango(location_code, 'redline', time_range=time_range)
    >>> fig = plt.figure(layout='constrained', figsize=(6, 6.5))
    >>> gs = matplotlib.gridspec.GridSpec(2, 1, fig, height_ratios=(3, 1))
    >>> ax = asilib.map.create_map(lat_bounds=(30, 45), lon_bounds=(-125, -100), fig_ax=(fig, gs[0]))
    >>> bx = fig.add_subplot(gs[1])
    >>> bx.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%H:%M'))
    >>> gen = asi.animate_map_gen(ax=ax, asi_label=True, overwrite=True)
    >>>
    >>> cdas = cdasws.CdasWs()
    >>> time_range = cdasws.TimeInterval(
    >>>     datetime.fromisoformat(str(time_range[0]-timedelta(days=0.5))).replace(tzinfo=timezone.utc),
    >>>     datetime.fromisoformat(str(time_range[1]+timedelta(days=0.5))).replace(tzinfo=timezone.utc)
    >>>     )
    >>> _, data = cdas.get_data(
    >>>                 'OMNI_HRO_5MIN', ['SYM_H'], time_range
    >>>                 )
    >>> symh = pd.DataFrame(index=data['SYM_H'].Epoch.data, data={'SYM_H':data['SYM_H']})
    >>> bx.plot(symh.index, symh['SYM_H'], c='k')
    >>> bx.set(xlabel='Time [HH:MM]', ylabel='Sym-H [nT]')
    >>>
    >>> for image_time, image, _, im in gen:
    >>>     # Add your code that modifies each image here...
    >>>     # To demonstrate, lets annotate each frame with the timestamp.
    >>>     # We will need to delete the prior text object, otherwise the current one
    >>>     # will overplot on the prior one---clean up after yourself!
    >>>     if 'text_obj' in locals():
    >>>         # text_obj.remove()  # noqa: F821
    >>>         _time_guide.remove()  # noqa: F821
    >>>     text_obj = plt.suptitle(f'MANGO-{location_code} | {image_time:%F %T}', fontsize=15)
    >>>     _time_guide = bx.axvline(image_time, c='k', ls='--')  
    
    Returns
    -------
    :py:meth:`~asilib.imager.Imager`
        A MANGO ASI instance with the time stamps, images, skymaps, and metadata.
    """
    if time is not None:
        time = utils.validate_time(time)
    else:
        time_range = utils.validate_time_range(time_range)

    local_image_dir = local_base_dir / 'images' / location_code.lower()

    channel = channel.lower()
    assert channel[0] in ['r', 'g'], (f"{channel} is an invalid MANGO color channel. "
                                      f"Try either 'redline' or 'greenline'.")
    if channel[0] == 'r':
        channel = 'redline'
    else:
        channel = 'greenline'

    mango_site_info = mango_info()
    mango_site_info = mango_site_info[
        mango_site_info['location_code'] == location_code.upper()
        ].reset_index()
    assert mango_site_info.shape[0] == 1, (
        f"MANGO-{location_code.upper()} is an invalid location. Try one of these: "
        f"{list(mango_info()['location_code'])}.")

    file_paths = _get_image_files(
        location_code,
        channel,
        time,
        time_range,
        base_url,
        local_image_dir,
        redownload,
        missing_ok,
    )

    start_times = len(file_paths) * [None]
    end_times = len(file_paths) * [None]
    for i, file_path in enumerate(file_paths):
        date_match = re.search(r'\d{8}', file_path.name)
        start_times[i] = datetime.strptime(date_match.group(), '%Y%m%d')
        end_times[i] = start_times[i] + timedelta(days=1)
    file_info = {
        'path': file_paths,
        'start_time': start_times,
        'end_time': end_times,
        'loader': _load_h5,
    }
    if time_range is not None:
        file_info['time_range'] = time_range
        if len(file_paths) == 0:
            raise ValueError(f'No MANGO-{location_code.upper()} files found for {file_info["time_range"]}')
    else:
        file_info['time'] = time
        if len(file_paths) == 0:
            raise ValueError(f'No MANGO-{location_code.upper()} files found for {time}')
    

    mango_meta = _load_h5_meta(file_paths[0])
    
    meta = {
        'array': 'MANGO',
        'channel':channel,
        'location': location_code.upper(),
        'lat': mango_site_info.loc[0, 'latitude'],
        'lon': mango_site_info.loc[0, 'longitude'],
        'alt': None,
        'cadence': mango_meta['cadence'],
        'resolution': mango_meta['resolution'],
        'acknowledgment': (
            'The imaging data are obtained through the MANGO network and operated by SRI with '
            'support from US National Science Foundation award AGS-1933013. Please reach out '
            'to the PI Asti Bhatt (asti.bhatt@sri.com) before using the data for publication. '
            'Cite https://doi.org/10.1029/2023JA031589 for the description of the MANGO network.'
            )
    }

    plot_settings = {
        'color_map': matplotlib.colors.LinearSegmentedColormap.from_list(
            'black_to_red', ['k', channel[0]]
            ),
        'color_bounds':(50, 250)
    }

    skymap = {
        'lat':mango_meta['lat'],
        'lon':mango_meta['lon'],
        'az':mango_meta['az'],
        'el':mango_meta['el'],
        'alt':mango_meta['alt'],
        'path':mango_meta['path']
    }
    if acknowledge and ('mango' not in asilib.config['ACKNOWLEDGED_ASIS']):
        print(meta['acknowledgment'])
        asilib.config['ACKNOWLEDGED_ASIS'].append('mango')
    return imager(file_info, meta, skymap, plot_settings=plot_settings)


def mango_info() -> pd.DataFrame:
    """
    Returns a pd.DataFrame with the MANGO ASI names and locations.

    Returns
    -------
    pd.DataFrame
        A table of THEMIS imager names and locations.
    """
    path = pathlib.Path(asilib.__file__).parent / 'data' / 'asi_locations.csv'
    df = pd.read_csv(path)
    df = df[df['array'] == 'MANGO']
    return df.reset_index(drop=True)

def _get_image_files(
    location_code: str,
    channel:str,
    time: datetime,
    time_range: Iterable[datetime],
    base_url: str,
    save_dir: Union[str, pathlib.Path],
    redownload: bool,
    missing_ok: bool,
    ) -> List[pathlib.Path]:
    """
    Find MANGO image files either locally or download them from the internet.

    Parameters
    ----------
    location_code:str
        The three-letter ASI location code.
    channel: str
        The color channel. Could be "redline", "greenline", "r", or "g".
    time: datetime.datetime
        Time to download one file. Either time or time_range must be specified,
        but not both.
    time_range: Iterable[datetime]
        An iterable with a start and end time. Either time or time_range must be
        specified, but not both.
    base_url: str
        The starting URL to search for file.
    save_dir: str or pathlib.Path
        The parent directory where to save the data to.
    redownload: bool
        Download data even if the file is found locally. This is useful if data
        is corrupt.
    missing_ok: bool
        Wether to allow missing data files inside time_range (after searching
        for them locally and online).

    Returns
    -------
    list(pathlib.Path)
        Local paths to each h5 file that was successfully found.
    """
    if (time is None) and (time_range is None):
        raise ValueError('time or time_range must be specified.')
    elif (time is not None) and (time_range is not None):
        raise ValueError('both time and time_range can not be simultaneously specified.')
    
    if redownload:
        # Option 1/4: Download one minute of data regardless if it is already saved
        if time is not None:
            return [
                _download_one_file(
                    location_code, channel, time, base_url, save_dir, redownload
                        )
            ]

        # Option 2/4: Download the data in time range regardless if it is already saved.
        elif time_range is not None:
            time_range = utils.validate_time_range(time_range)
            file_times = utils.get_filename_times(time_range, dt='days')
            file_paths = []
            for file_time in file_times:
                try:
                    file_paths.append(
                        _download_one_file(
                            location_code, channel, file_time, base_url, save_dir, redownload
                        )
                    )
                except (FileNotFoundError, AssertionError) as err:
                    if missing_ok and (
                        ('does not contain any hyper references containing' in str(err))
                        or ('Only one href is allowed' in str(err))
                    ):
                        continue
                    raise
            return file_paths
    else:
        # Option 3/4: Download one minute of data if it is not already saved.
        if time is not None:
            file_search_str = f'mango-{location_code.lower()}-{channel}-level1-{time:%Y%m%d}.hdf5'
            local_file_paths = list(pathlib.Path(save_dir).rglob(file_search_str))
            if len(local_file_paths) == 1:
                return local_file_paths
            else:
                return [
                    _download_one_file(
                        location_code, channel, time, base_url, save_dir, redownload
                    )
                ]

        # Option 4/4: Download the data in time range if they don't exist locally.
        elif time_range is not None:
            time_range = utils.validate_time_range(time_range)
            file_times = utils.get_filename_times(time_range, dt='days')
            file_paths = []
            for file_time in file_times:
                file_search_str = (
                    f'mango-{location_code.lower()}-{channel}-level1-{file_time:%Y%m%d}.hdf5'
                )
                local_file_paths = list(pathlib.Path(save_dir).rglob(file_search_str))
                if len(local_file_paths) == 1:
                    file_paths.append(local_file_paths[0])
                else:
                    try:
                        file_paths.append(
                            _download_one_file(
                                location_code, channel, file_time, base_url, save_dir, redownload
                            )
                        )
                    except (FileNotFoundError, AssertionError, requests.exceptions.HTTPError) as err:
                        if missing_ok and (
                            ('does not contain any hyper references containing' in str(err)) or
                            ('Only one href is allowed' in str(err)) or
                            ('404 Client Error: Not Found for url:' in str(err))
                        ):
                            continue
                        raise
            if missing_ok and len(file_paths) == 0:
                if time_range is not None:
                    warnings.warn(
                        f'No local or online image files found for MANGO-{location_code.upper()} '
                        f'for {time_range=}.'
                        )
                else:
                    warnings.warn(
                        f'No local or online image files found for MANGO-{location_code.upper()} '
                        f'for {time=}.'
                        )
            return file_paths
    return

def _download_one_file(
    location_code: str,
    channel:str,
    time: datetime,
    base_url: str,
    save_dir: Union[str, pathlib.Path],
    redownload: bool,
) -> pathlib.Path:
    """
    Download one h5 file.

    Parameters
    ----------
    location_code: str
        The ASI three-letter location code.
    channel: str
        The ASI channel
    time: str or datetime.datetime
        A time to look for the ASI data at.
    base_url: str
        The starting URL to search for file.
    save_dir: str or pathlib.Path
        The parent directory where to save the data to.
    redownload: bool
        Will redownload an existing file.

    Returns
    -------
    pathlib.Path
        Local path to file.
    """
    start_url = base_url + f'{location_code.lower()}/{channel}/level1/{time.year}/{time:%j}/'
    file_search_str = f'mango-{location_code.lower()}-{channel}-level1-{time:%Y%m%d}.hdf5'
    d = download.Downloader(start_url)
    matched_downloaders = d.ls(file_search_str)
    assert len(matched_downloaders) == 1
    return matched_downloaders[0].download(save_dir, redownload=redownload)


def _load_h5(file_path):
    with h5py.File(file_path, 'r') as file:
        # Assigning timezone.utc and immediately removing it is necessary to correctly convert 
        # to the UTC datetime objects. 
        start_exposure_times = np.array([
            datetime.fromtimestamp(ti, tz=timezone.utc).replace(tzinfo=None) 
            for ti in file['UnixTime'][0, :]
            ])
        images = file['ImageData'][...].astype(float)
        images[np.where(images == 0)] = np.nan
        return start_exposure_times, images

def _load_h5_meta(file_path):
    with h5py.File(file_path, 'r') as file:
        dt = file['UnixTime'][0, 1:]-file['UnixTime'][0, :-1]
        assert np.all(dt-dt[0] < dt), (
            f'There is a timestamp jump in file {file_path}. {min(dt)}<dt<{max(dt)}.'
            )
        meta_dict = {
            'lon':file['Longitude'][:], 
            'lat':file['Latitude'][:],
            'az':file['Azimuth'][:],
            'el':file['Elevation'][:],
            'alt':None,  # TODO ask Leslie what altitude they mapped to.
            'resolution':file['ImageData'].shape[1:],
            'cadence':dt[0],
            'path':file_path
        }
        return meta_dict
    
if __name__ == '__main__':
    import cdasws
    import matplotlib.pyplot as plt
    import matplotlib.dates

    time_range=(datetime(2021, 11, 4, 1, 0), datetime(2021, 11, 4, 12, 24))
    location_code='CFS'
    asi = mango(location_code, 'redline', time_range=time_range)

    fig = plt.figure(layout='constrained', figsize=(6, 6.5))
    gs = matplotlib.gridspec.GridSpec(2, 1, fig, height_ratios=(3, 1))
    ax = asilib.map.create_map(lat_bounds=(30, 45), lon_bounds=(-125, -100), fig_ax=(fig, gs[0]))
    bx = fig.add_subplot(gs[1])
    bx.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%H:%M'))

    gen = asi.animate_map_gen(ax=ax, asi_label=True, overwrite=True)

    cdas = cdasws.CdasWs()
    time_range = cdasws.TimeInterval(
        datetime.fromisoformat(str(time_range[0]-timedelta(days=0.5))).replace(tzinfo=timezone.utc), 
        datetime.fromisoformat(str(time_range[1]+timedelta(days=0.5))).replace(tzinfo=timezone.utc)
        )
    _, data = cdas.get_data(
                    'OMNI_HRO_5MIN', ['SYM_H'], time_range
                    )
    symh = pd.DataFrame(index=data['SYM_H'].Epoch.data, data={'SYM_H':data['SYM_H']})
    bx.plot(symh.index, symh['SYM_H'], c='k')
    bx.set(xlabel='Time [HH:MM]', ylabel='Sym-H [nT]')

    for image_time, image, _, im in gen:
        # Add your code that modifies each image here...
        # To demonstrate, lets annotate each frame with the timestamp.
        # We will need to delete the prior text object, otherwise the current one
        # will overplot on the prior one---clean up after yourself!
        if 'text_obj' in locals():
            # text_obj.remove()  # noqa: F821
            _time_guide.remove()  # noqa: F821
        text_obj = plt.suptitle(f'MANGO-{location_code} | {image_time:%F %T}', fontsize=15)
        _time_guide = bx.axvline(image_time, c='k', ls='--')


from datetime import datetime

import matplotlib.pyplot as plt

import asilib.asi

location_code = 'RANK'
time = datetime(2017, 9, 15, 2, 34, 0)

asi = asilib.asi.themis(location_code, time=time)
ax, im = asi.plot_fisheye()
plt.colorbar(im)
ax.axis('off')
plt.show()



"""
Create a world map showcasing the ASI arrays that asilib supports.
"""
from datetime import datetime

import matplotlib.pyplot as plt
import matplotlib.colors
import matplotlib.patches
import numpy as np
import requests
import cartopy.crs as ccrs
import cartopy.feature

import asilib.asi
from asilib.imager import Skymap_Cleaner

asi_arrays = {
    'THEMIS':(asilib.asi.themis, asilib.asi.themis_info, 'k', 10),
    'REGO':(asilib.asi.rego, asilib.asi.rego_info, 'blue', 10),
    'TREx-NIR':(asilib.asi.trex_nir, asilib.asi.trex_nir_info, 'c', 12),
    'TREx-RGB':(asilib.asi.trex_rgb, asilib.asi.trex_rgb_info, 'purple', 15),
    'PsA Project':(asilib.asi.psa_project, asilib.asi.psa_project_info, 'orange', 15),
    }

mango_sations = [
    ['CVO', datetime(2024, 10, 25, 5, 28), 'r'],
    ['CFS', datetime(2024, 10, 25, 5, 28), 'r'],
    ['MTO', datetime(2024, 10, 25, 5, 28), 'r'],
    ['EIO', datetime(2024, 10, 25, 5, 28), 'r'],
    ['PAR', datetime(2024, 10, 25, 5, 28), 'r'],
    ['MDK', datetime(2023, 10, 14, 6, 44), 'r']
]

time = datetime(2022, 1, 1)

fig = plt.figure(figsize=(8, 6))
ax = fig.add_subplot(projection=ccrs.Orthographic(-100, 50))

ax.add_feature(cartopy.feature .LAND, color='g')
ax.add_feature(cartopy.feature .OCEAN, color='w')
ax.add_feature(cartopy.feature .COASTLINE, edgecolor='k')
ax.gridlines()

plotted_locations = []

for mango_station in mango_sations:
    asi = asilib.asi.mango(mango_station[0], 'redline', time=mango_station[1])
    ax.contour(
        asi.skymap['lon'], asi.skymap['lat'], asi.skymap['el'], levels=[16],
        transform=ccrs.PlateCarree(), colors=mango_station[2]
        )
    if mango_station[0] not in plotted_locations:
        ax.text(
            asi.meta['lon'],
            asi.meta['lat'],
            asi.meta['location'].upper(),
            color='k',
            transform = ccrs.PlateCarree(),
            va='center',
            ha='center',
        )
        plotted_locations.append(mango_station[0])

ax.text(
    0, len(asi_arrays)/30, 'MANGO-redline', va='bottom', color=mango_station[2], 
    transform=ax.transAxes, fontsize=15
    )

for i, (array, (loader, info_df, color, elevation)) in enumerate(asi_arrays.items()):
    asi_array_info = info_df()
    for location in asi_array_info['location_code']:
        try:
            asi = loader(location, time=time, load_images=False)
        except requests.exceptions.HTTPError as err:
            if ('Not Found for url' in str(err)) or ('Precondition Failed for url' in str(err)):
                print(err)
                continue
            else:
                raise
        _skymap_cleaner = Skymap_Cleaner(
                asi.skymap['lon'], 
                asi.skymap['lat'], 
                asi.skymap['el'],
            )
        _lon, _lat = _skymap_cleaner.remove_nans()
        if (
            (_lon.shape[0] == asi.skymap['el'].shape[0]+1) and 
            (_lon.shape[1] == asi.skymap['el'].shape[1]+1)
            ):
            _lon = _lon[:-1, :-1]
            _lat = _lat[:-1, :-1]
        ax.contour(
            _lon, _lat, asi.skymap['el'], levels=[elevation],
            transform=ccrs.PlateCarree(), colors=color
        )
        if location not in plotted_locations:
            ax.text(
                asi.meta['lon'],
                asi.meta['lat'],
                asi.meta['location'].upper(),
                color='k',
                transform = ccrs.PlateCarree(),
                va='center',
                ha='center',
            )
            plotted_locations.append(location)
    ax.text(0, i/30, array, va='bottom', color=color, transform=ax.transAxes, fontsize=15)

ax.set_title(f'Imagers Supported by asilib as of {datetime.now().date()}', fontsize=22)        
plt.tight_layout()
plt.show()


from .rego import rego, rego_info, rego_skymap
from .themis import themis, themis_info, themis_skymap, themis_available, plot_themis_available
from .trex import trex_nir, trex_nir_skymap, trex_nir_info
from .trex import trex_rgb, trex_rgb_info, trex_rgb_skymap, trex_rgb_available, plot_trex_rgb_available
from .psa_project import psa_project, psa_project_info, psa_project_skymap, psa_project_lamp
from .mango import mango, mango_info

__all__ = [
    'themis', 
    'themis_info', 
    'themis_skymap',
    'themis_available',
    'plot_themis_available',
    'rego',
    'rego_info', 
    'rego_skymap',
    'trex_rgb', 
    'trex_rgb_info', 
    'trex_rgb_skymap',
    'trex_nir', 
    'trex_nir_skymap', 
    'trex_nir_info',
    'trex_rgb_available', 
    'plot_trex_rgb_available',
    'mango',
    'mango_info',
    'psa_project',
    'psa_project_info',
    'psa_project_lamp'
]  # So Sphinx shortens the name. See https://stackoverflow.com/a/31594545



"""
A set of utility functions for asilib.
"""
import dateutil.parser
import shutil
from typing import List, Union
from collections.abc import Iterable
import copy
from datetime import timedelta, datetime

import pandas as pd
import numpy as np


_time_type = Union[datetime, str]
_time_range_type = List[_time_type]


def validate_time(time: _time_type) -> List[datetime]:
    """
    Validates tries to parse the time into datetime objects.
    """
    if isinstance(time, str):
        time = dateutil.parser.parse(time)
    elif isinstance(time, (datetime, pd.Timestamp)):
        pass
    else:
        raise ValueError(f'Unknown time format, {time}')
    _validate_year(time)
    return time


def validate_time_range(time_range: _time_range_type) -> List[datetime]:
    """
    Validates tries to parse the time_range into datetime objects.
    """
    if time_range is None:
        return None

    assert isinstance(
        time_range, (list, tuple, np.ndarray)
    ), "time_range must be a list, tuple, or np.ndarray."
    assert len(time_range) == 2, "time_range must be a list or a tuple with start and end times."

    time_range_parsed = []

    for t in time_range:
        if isinstance(t, str):
            time_range_parsed.append(dateutil.parser.parse(t))
        elif isinstance(t, (int, float)):
            raise ValueError(f'Unknown time format, {t}')
        else:
            time_range_parsed.append(t)

    for t in time_range_parsed:
        _validate_year(t)

    time_range_parsed.sort()
    return time_range_parsed


def _validate_year(time):
    """
    Provides a sanity check that the year is after 2000 and before the current year + 1.
    """
    year = time.year
    if year < 2000:
        raise ValueError(f'The passed year={year} must be greater than 2000.')
    elif year > datetime.now().year:
        raise ValueError(f'The passed year={year} must be less than the current year + 1.')
    return


def get_filename_times(time_range: _time_range_type, dt='hours') -> List[datetime]:
    """
    Returns the dates and times within time_range and in dt steps. It returns times
    with the smaller components than dt set to zero, and increase by dt. This is useful
    to create a list of dates and times to load sequential files.

    time_range: list[datetime or str]
        A start and end time to calculate the file dates.
    dt: str
        The time difference between times. Can be 'days', 'hours', or 'minutes'.
    """
    time_range = validate_time_range(time_range)
    assert dt in ['days', 'hours', 'minutes'], "dt must be 'day', 'hour', or 'minute'."
    # First we need to appropriately zero time_range[0] so that the file that contains the
    # time_range[0] is returned. For example, if dt='hour' and time_range[0] is not at the
    # top of the hour, we zero the smaller time components. So
    # time_range[0] = '2010-01-01T10:29:32.000000' will be converted to '2010-01-01T10:00:00.000000'
    # to allow the
    zero_time_chunks = {'microsecond': 0, 'second': 0}
    # We don't need an if-statement if dt == 'minute'
    if dt == 'hours':
        zero_time_chunks['minute'] = 0
    elif dt == 'days':
        zero_time_chunks['minute'] = 0
        zero_time_chunks['hour'] = 0
    current_time = copy.copy(time_range[0].replace(**zero_time_chunks))
    times = []

    # Not <= in while loop because we don't want to download the final time if time_range[1] is,
    # exactly matches the end file name (you don't want to download the 11th hour if
    # time_range[1] is 'YYY-MM-DDT11:00:00').
    while current_time < time_range[1]:
        times.append(current_time)
        current_time += timedelta(**{dt: 1})
    return times


def progressbar(iterator: Iterable, iter_length: int = None, text: str = None):
    """
    A terminal progress bar.

    Parameters
    ----------
    iterator: Iterable
        The iterable that will be looped over.
    iter_length: int
        How many items the iterator loop over. If None, will calculate it
        using len(iterator).
    text: str
        Insert an optional text string in the beginning of the progressbar.
    """
    if text is None:
        text = ''
    else:
        text = text + ':'

    if iter_length is None:
        iter_length = len(iterator)

    try:
        for i, item in enumerate(iterator):
            i += 1  # So we end at 100%. Happy users!
            terminal_cols = shutil.get_terminal_size(fallback=(80, 20)).columns
            max_cols = int(terminal_cols - len(text) - 10)
            # Prevent a crash if the terminal window is narrower then len(text).
            if max_cols < 0:
                max_cols = 0

            percent = round(100 * i / iter_length)
            bar = "#" * int(max_cols * percent / 100)
            print(f'{text} |{bar:<{max_cols}}| {percent}%', end='\r')
            yield item
    finally:
        print()  # end with a newline.

def stretch_contrast(image, vmin, vmax):
    """
    This function performs Normalization, also called Contrast Stretching to an image
    given min/max values. For more information see 
    https://en.wikipedia.org/wiki/Normalization_(image_processing), and
    https://stackoverflow.com/a/71835553.

    To not apply this algorithm use vmin=0 and vmax=256 for 8-bit images (i.e., TREx).
    """
    image = (image-vmin)/(vmax-vmin)
    # Above is a good enough scaling, but there may be some >1 values since vmin, vmax 
    # are percentiles (it better handles cases when part of the image is saturated by 
    # moonlight, for example).
    image[image > 1] = 1
    image[image < 0] = 0
    return image


"""
Recreate the auroral arc that is shown in Fig. 2b in:
Donovan, E., Liu, W., Liang, J., Spanswick, E., Voronkov, I., Connors, M., ... & Rae,
I. J. (2008). Simultaneous THEMIS in situ and auroral observations of a small substorm.
Geophysical Research Letters, 35(17).
"""

from datetime import datetime

import matplotlib.pyplot as plt

import asilib
import asilib.map
import asilib.asi

time = datetime(2007, 3, 13, 5, 8, 45)
location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
map_alt = 110
min_elevation = 2

ax = asilib.map.create_map(lon_bounds=(-140, -60), lat_bounds=(40, 82))

_imagers = []

for location_code in location_codes:
    _imagers.append(asilib.asi.themis(location_code, time=time, alt=map_alt))

asis = asilib.Imagers(_imagers)
asis.plot_map(ax=ax, overlap=False, min_elevation=min_elevation)

ax.set_title('Donovan et al. 2008 | First breakup of an auroral arc')
plt.show()



from datetime import datetime

import asilib.asi

location_code = 'FSMI'
time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 30))
asi = asilib.asi.themis(location_code, time_range=time_range)
asi.animate_fisheye()

print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')



"""
Tests for the asilib.Imagers class.
"""
import matplotlib.testing.decorators
import numpy as np


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_donovan_arc_example'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_donovan_arc_example():
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib
    import asilib.map
    import asilib.asi
    time = datetime(2007, 3, 13, 5, 8, 45)
    location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
    map_alt = 110
    min_elevation = 2
    ax = asilib.map.create_simple_map(lon_bounds=(-140, -60), lat_bounds=(40, 82))
    _imagers = []
    for location_code in location_codes:
        _imagers.append(asilib.asi.themis(location_code, time=time, alt=map_alt))
    asis = asilib.Imagers(_imagers)
    asis.plot_map(ax=ax, overlap=False, min_elevation=min_elevation)
    ax.set_title('Donovan et al. 2008 | First breakup of an auroral arc')
    return

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_bad_image'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_plot_bad_image():
    """
    There is a bad image in TREX-RGB data taken at PINA. This test will fail
    once the data file is fixed, and this tests the asilib.Imagers() initialization
    with a single asilib.Imager() instance.
    """
    import matplotlib.pyplot as plt

    import asilib
    import asilib.asi

    asi = asilib.asi.trex_rgb('PINA', time='2021-11-04T05:25:00')
    asis = asilib.Imagers(asi)
    _, ax = plt.subplots()
    asis.plot_fisheye(ax)
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_fisheye'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_plot_fisheye():
    from datetime import datetime

    import matplotlib.pyplot as plt
    import asilib
    import asilib.asi

    time = datetime(2007, 3, 13, 5, 8, 45)
    location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']

    fig, ax = plt.subplots(1, len(location_codes), figsize=(12, 3.5))

    _imagers = []

    for location_code in location_codes:
        _imagers.append(asilib.asi.themis(location_code, time=time))

    for ax_i in ax:
        ax_i.axis('off')

    asis = asilib.Imagers(_imagers)
    asis.plot_fisheye(ax=ax)

    plt.suptitle('Donovan et al. 2008 | First breakup of an auroral arc')
    plt.tight_layout()
    return


def test_get_points():
    from datetime import datetime
    
    import asilib
    import asilib.asi
    
    time = datetime(2007, 3, 13, 5, 8, 45)
    location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
    map_alt = 110
    min_elevation = 2
    
    _imagers = []
    
    for location_code in location_codes:
        _imagers.append(asilib.asi.themis(location_code, time=time, alt=map_alt))
    
    asis = asilib.Imagers(_imagers)
    lat_lon_points, intensities = asis.get_points(min_elevation=min_elevation)
    assert lat_lon_points.shape == (174857, 2)
    assert intensities.shape == (174857,)
    np.testing.assert_almost_equal(
        lat_lon_points[:10, :],
        np.array([
            [  54.35538483, -126.76580811],
            [  54.34091568, -126.89642334],
            [  54.32193756, -127.0307312 ],
            [  54.29838562, -127.16894531],
            [  54.27019119, -127.31124878],
            [  54.23726654, -127.45788574],
            [  54.19952011, -127.60906982],
            [  54.15684509, -127.76501465],
            [  54.10913086, -127.92602539],
            [  54.0562439 , -128.09228516]
            ])
    )
    np.testing.assert_almost_equal(
        lat_lon_points[-10:, :],
        np.array([
            [ 64.15480804, -82.1703186 ],
            [ 64.12185669, -82.30700684],
            [ 64.09317017, -82.44506836],
            [ 64.06869507, -82.58477783],
            [ 64.04837036, -82.72647095],
            [ 64.03216553, -82.87042236],
            [ 64.02003479, -83.01693726],
            [ 64.01195526, -83.16635132],
            [ 64.00791168, -83.31893921],
            [ 64.00789642, -83.47509766]
            ])
       )
    np.testing.assert_equal(
        intensities[:10],
        np.array([2823., 2752., 2738., 2738., 2746., 2702., 2685., 2616., 2592., 2569.])
        )
    return

def test_iterate_trex_imagers():
    """
    Test the TREx imagers (time, image) iterations are in sync, and if not, that the image is 
    correctly masked. 
    """
    import asilib
    import asilib.asi

    time_range = ('2023-02-24T05:10', '2023-02-24T05:15')
    # Load all TREx imagers.
    trex_metadata = asilib.asi.trex_rgb_info()
    asis = asilib.Imagers(
        [asilib.asi.trex_rgb(location_code, time_range=time_range) 
        for location_code in trex_metadata['location_code']],
        sync_image_tol=0.5
        )

    _guide_times = []
    _times = []
    _images = [] 
    for _guide_time, _asi_times, _asi_images in asis:
        _guide_times.append(_guide_time)
        _times.append(_asi_times)
        _images.append(_asi_images)

    dt = np.full(np.array(_times).shape, np.nan)
    for i, (_guide_time, imager_times_i) in enumerate(zip(_guide_times, _times)):
        dt[i, :] = [(_guide_time-j).total_seconds() for j in imager_times_i]
    
    dt[np.abs(dt) > asis.imagers[0].meta['cadence']] = np.nan

    assert np.nanmax(np.abs(dt)) == 0.358808  # Maximum synchronized time difference.
    assert np.sum(np.isnan(dt)) <= 1 # One image is missing and thus masked.
    return

def test_iterate_rego_imagers():
    """
    Test the REGO imagers (time, image) iterations are in sync, and if not, that the image is 
    correctly masked. 

    Event from Panov+2019 https://doi.org/10.1029/2019JA026521
    """
    import asilib
    import asilib.asi

    time_range = ('2016-08-09T08:00', '2016-08-09T08:05')
    
    asis = asilib.Imagers(
        [asilib.asi.rego(location_code, time_range=time_range) 
        for location_code in ['GILL', 'FSMI', 'FSIM']]
        )

    _guide_times = []
    _times = []
    _images = [] 
    for _guide_time, _asi_times, _asi_images in asis:
        _guide_times.append(_guide_time)
        _times.append(_asi_times)
        _images.append(_asi_images)

    dt = np.full(np.array(_times).shape, np.nan)
    for i, (_guide_time, imager_times_i) in enumerate(zip(_guide_times, _times)):
        dt[i, :] = [(_guide_time-j).total_seconds() for j in imager_times_i]

    assert np.max(np.abs(dt)) == 0
    
    # 100 time stamps, 3 imagers, and x- and y- pixels for each image.
    # This will fail if any images are labeled as None. The error is: 
    # "ValueError: setting an array element with a sequence. The requested 
    # array has an inhomogeneous shape after 2 dimensions. The detected 
    # shape was (100, 3) + inhomogeneous part.
    assert np.array(_images).shape == (100, 3, 512, 512)
    return

def test_iterate_themis_imagers():
    """
    Test the THEMIS imagers (time, image) iterations are in sync, and if not, that the image is 
    correctly masked. 

    Event from Panov+2019 https://doi.org/10.1029/2019JA026521
    """
    from datetime import datetime

    import asilib
    import asilib.asi

    time_range = (
        datetime(2007, 3, 13, 5, 5, 0),
        datetime(2007, 3, 13, 5, 10, 0)
        )
    
    asis = asilib.Imagers(
        [asilib.asi.themis(location_code, time_range=time_range) 
        for location_code in ['FSIM', 'ATHA', 'TPAS', 'SNKQ']]
        )

    _guide_times = []
    _times = []
    _images = [] 
    for _guide_time, _asi_times, _asi_images in asis:
        _guide_times.append(_guide_time)
        _times.append(_asi_times)
        _images.append(_asi_images)

    dt = np.full(np.array(_times).shape, np.nan)
    for i, (_guide_time, imager_times_i) in enumerate(zip(_guide_times, _times)):
        dt[i, :] = [(_guide_time-j).total_seconds() for j in imager_times_i]

    assert np.max(np.abs(dt)) == 0.075671
    
    # 100 time stamps, 4 imagers, and x- and y- pixels for each image.
    # This will fail if any images are labeled as None. The error is: 
    # "ValueError: setting an array element with a sequence. The requested 
    # array has an inhomogeneous shape after 2 dimensions. The detected 
    # shape was (100, 3) + inhomogeneous part.
    assert np.array(_images).shape == (100, 4, 256, 256)
    return

def test_animate_map_sync_bug():
    """
    Test Imagers.animate_map() and ensure that the images are all synchronized.
    """
    import asilib
    import asilib.asi

    asi_list = [asilib.asi.themis(location, time_range=('2008-02-11T03:35', '2008-02-11T03:36'))
                for location in ['ATHA', 'FSIM', 'FSMI']]
    asis = asilib.Imagers(asi_list)
    asis.animate_map(overwrite=True)
    return

def test_animate_map():
    """
    Just check that if it crashes or not.
    """
    import asilib
    import asilib.asi

    time_range = ('2021-11-04T06:59:50', '2021-11-04T07:00:50')
    asis = asilib.Imagers(
        [asilib.asi.trex_rgb(location_code, time_range=time_range) 
        for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
        )
    asis.animate_map(lon_bounds=(-115, -85), lat_bounds=(43, 63), overwrite=True)
    return

def test_multi_asi_iter():
    """
    Loop over multiple ASI imagers and ensure that the time differences between the guide time
    and each imager time are within the cadence of each imager.
    """
    from datetime import datetime
    import asilib.asi

    time_range=(datetime(2021, 11, 4, 1, 56), datetime(2021, 11, 4, 2, 2))
    mango_location_code='CFS'
    mango_asi = asilib.asi.mango(mango_location_code, 'redline', time_range=time_range)
    trex_location_code = 'FSMI'
    trex_asi = asilib.asi.trex_rgb(trex_location_code, time_range=time_range)
    asis = asilib.Imagers([mango_asi, trex_asi])

    mango_dt_within_cadence = []
    trex_dt_within_cadence = []

    for guide_time, _asi_times, _ in asis:
        mango_dt_within_cadence.append(
            mango_asi.meta['cadence'] > (guide_time - _asi_times[0]).total_seconds()
            )
        trex_dt_within_cadence.append(
            trex_asi.meta['cadence'] > (guide_time - _asi_times[1]).total_seconds()
            )
    assert all(mango_dt_within_cadence)
    assert all(trex_dt_within_cadence)
    return

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_pixel_overlap'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_overlap_pixels():
    """
    Test the Imagers.find_overlap_pixels() pixels and plot them on a map.
    """
    from datetime import datetime

    import asilib.asi
    import asilib.map
    import matplotlib.pyplot as plt

    time = '2021-11-04T06:55'
    min_elevation = 10
    asis = asilib.Imagers(
        [asilib.asi.trex_rgb(location_code, time=time) 
        for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
        )
    overlapping_masks = asis.find_overlap_pixels(min_elevation=min_elevation)

    ax = asilib.map.create_simple_map(
        lon_bounds=(-120, -80), lat_bounds=(40, 70)
        )            

    asis.plot_map(ax=ax, min_elevation=min_elevation)

    asi_names = [f'{_imager.meta["array"]}-{_imager.meta["location"]}' for _imager in asis.imagers]

    n = 0
    for _, inner_dict in overlapping_masks.items():
        n += len(inner_dict)
    colors = iter(plt.cm.viridis(np.linspace(0,1,n)))

    for self_loc, neighbors in overlapping_masks.items():
        for neighbor_loc, overlapping_mask in neighbors.items():
            self_loc_idx = asi_names.index(self_loc)
            # neighbor_loc_idx = asi_names.index(neighbor_loc)
            ax.scatter(
                asis.imagers[self_loc_idx].skymap['lon'][overlapping_mask], 
                asis.imagers[self_loc_idx].skymap['lat'][overlapping_mask], 
                c=next(colors),
                alpha=0.1, 
                label=f'{self_loc} pixels overlapping with {neighbor_loc}',
                s=10
            )
    plt.tight_layout()
    return

# @matplotlib.testing.decorators.image_comparison(
#     baseline_images=['test_plot_map_eq'],
#     tol=10,
#     remove_text=True,
#     extensions=['png'],
# )
# def test_plot_map_eq():
#     """
#     Test the Imagers.plot_map_eq() method using IGRF.
#     """
#     import asilib
#     import asilib.asi
#     import matplotlib.pyplot as plt

#     time = '2021-11-04T06:59'
#     location_codes = ['GILL']
#     asis = asilib.Imagers(
#         [asilib.asi.trex_rgb(location_code, time=time) 
#         for location_code in location_codes]
#         )
#     fig, ax = plt.subplots()
#     asis.plot_map_eq(ax=ax, b_model="IGRF")
#     return


from datetime import datetime, timedelta
from typing import Tuple

import numpy as np
import matplotlib.pyplot as plt

import asilib.asi.themis


def footprint(
    center_lon: float,
    alt: float = 110,
    obit_period: float = 95 * 60,
    precession_rate: float = 10,
    cadence: float = 6,
    time_range: tuple = (datetime(2015, 1, 1), datetime(2015, 1, 2)),
) -> Tuple[np.array, np.array]:
    """
    Create a fake ephemeris for a LEO satellite.

    Parameters
    ----------
    center_lon:float
        The longitude at the midpoint of the time range.
    alt:float
        The footprint altitude in kilometers.
    obit_period:float
        The orbit period in seconds.
    precession_rate:float
        Controls how much the longitude change during the time_range,
        with the center_lon in the middle times.
    cadence:float
        The cadence of the footprint timestamps in seconds.
    time_range:tuple
        The start and end times that bound the footprint path.

    Returns
    -------
    np.array
        Footprint timestamps in datetime.datetime format.
    np.array
        An array of floats shaped (n_times, 3) with the footprint's
        lat, lon, alt coordinates.
    """
    n_seconds = (time_range[1] - time_range[0]).total_seconds()
    n_time_stamps = int(n_seconds / cadence)
    times = np.array([time_range[0] + timedelta(seconds=i * cadence) for i in range(n_time_stamps)])
    seconds_from_start = np.array([(time - times[0]).total_seconds() for time in times])

    lats = 90 * np.sin(2 * np.pi * seconds_from_start / obit_period)
    lons = np.linspace(
        center_lon - precession_rate / 2, center_lon + precession_rate, num=n_time_stamps
    )
    alts = alt * np.ones(n_time_stamps)
    lla = np.stack((lats, lons, alts)).T
    return times, lla


def visualize_footprint():
    """
    Visualize and validate that the footprint lons and lats are correct.
    """
    asi = asilib.asi.themis.themis('GILL', time='2015-01-01', load_images=False)
    times, lla = footprint(asi.meta['lon'])

    fig, ax = plt.subplots(3, 1)
    ax[0].plot(times, lla[:, 0])
    ax[1].plot(times, lla[:, 1])
    ax[2].plot(lla[:, 1], lla[:, 0])

    ax[0].set(xlabel='time', ylabel='lat [deg]')
    ax[1].set(xlabel='time', ylabel='lon [deg]')
    ax[2].set(xlabel='lon [deg]', ylabel='lat [deg]')
    return


if __name__ == '__main__':
    visualize_footprint()
    plt.show()



"""
Tests the psa_project() data loading and the example plotting functions.
"""
from datetime import datetime

import numpy
import pandas
import matplotlib
import pytest
import matplotlib.testing.decorators

import asilib.asi

def test_valid_camera_code():
    location_code = 'c2'
    asi = asilib.asi.psa_project(
        location_code, 
        time=datetime(2019, 3, 1, 18, 30, 0),
        load_images=False
    )
    assert asi.meta['location'] == 'C2'
    return

def test_invalid_camera_code():
    location_code = 'C100'
    with pytest.raises(ValueError) as excinfo:
        asi = asilib.asi.psa_project(
            location_code, 
            time=datetime(2019, 3, 1, 18, 30, 0)
        )
    assert f"{location_code=} is invalid." in str(excinfo.value)
    return

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_fisheye'], 
    tol=10, 
    remove_text=True, 
    extensions=['png']
)
def test_plot_fisheye():
    """
    https://ergsc.isee.nagoya-u.ac.jp/psa-gnd/bin/psa.cgi?year=2017&month=03&day=07&jump=Plot
    """
    asi = asilib.asi.psa_project(
       'C1', 
       time=datetime(2017, 3, 7, 19, 35, 0),
       redownload=False
       )
    asi.plot_fisheye(origin=(0.9, 0.1))
    return


"""
Tests the rego() data loading and the example plotting functions.
"""
from datetime import datetime

import requests
import pytest
import matplotlib.testing.decorators

##########################################
############# TEST LOADERS ###############
##########################################


# def test_rego_time():
#     """
#     Tests that one file is loaded and themis() returns the correct file.
#     """
#     # Calls the download function
#     img = themis.themis('gill', time='2014-05-05T04:49:10', redownload=True)
#     # _data should not be accessed by the user.
#     assert img._data['time'] == datetime(2014, 5, 5, 4, 49, 9, 37070)
#     # And data() should be
#     assert img.data.time == datetime(2014, 5, 5, 4, 49, 9, 37070)
#     assert img.data[0] == datetime(2014, 5, 5, 4, 49, 9, 37070)
#     assert img.data.image[0, 0] == 3464
#     assert img.data[1][0, 0] == 3464
#     assert img.data.image[-1, -1] == 3477
#     assert img.skymap['path'].name == 'themis_skymap_gill_20130103-%2B_vXX.sav'
#     # Does not call the download function
#     img2 = themis.themis('gill', time='2014-05-05T04:49:10', redownload=False)
#     assert img2._data['time'] == datetime(2014, 5, 5, 4, 49, 9, 37070)
#     return


# def test_themis_time_range():
#     """
#     Tests that multiple files are loaded using themis()'s time_range kwarg.
#     """
#     img = themis.themis('gill', time_range=['2014-05-05T05:10', '2014-05-05T05:12'])
#     assert img._data['path'][0].name == '20140505_0510_gill_themis19_full.pgm.gz'
#     assert img._data['path'][1].name == '20140505_0511_gill_themis19_full.pgm.gz'

#     assert img.data.time[0] == datetime(2014, 5, 5, 5, 10, 0, 30996)
#     assert img.data.time[-1] == datetime(2014, 5, 5, 5, 11, 57, 23046)
#     assert img.data.time.shape == (40,)
#     assert img.data.image.shape == (40, 256, 256)
#     return


# def test_themis_no_file():
#     """
#     Tests that themis() returns a FileNotFound error when we try to download
#     non-existant data.
#     """
#     with pytest.raises(requests.exceptions.HTTPError):
#         themis.themis('pina', time='2011/07/07T02:00')
#     return


# def test_themis_partial_files():
#     """
#     Tests that themis() returns a FileNotFound error when we try to download
#     non-existant data.
#     """
#     with pytest.raises(FileNotFoundError):
#         themis.themis('pina', time_range=['2011/07/07T04:15', '2011/07/07T04:23'], missing_ok=False)

#     img = themis.themis(
#         'pina',
#         time_range=['2011/07/07T04:15', '2011/07/07T04:23:30'],
#         missing_ok=True,
#         redownload=True,
#     )
#     assert img._data['path'][0].name == '20110707_0421_pina_themis18_full.pgm.gz'
#     assert img._data['path'][1].name == '20110707_0422_pina_themis18_full.pgm.gz'
#     assert img._data['path'][2].name == '20110707_0423_pina_themis18_full.pgm.gz'
#     return


# def test_themis_asi_meta():
#     """
#     Checks that the THEMIS ASI metadata is correct.
#     """
#     img = themis.themis(
#         'pina',
#         time_range=['2011/07/07T04:20', '2011/07/07T04:22:00'],
#         missing_ok=True,
#         redownload=False,
#     )
#     assert img.meta == {
#         'array': 'THEMIS',
#         'location': 'PINA',
#         'lat': 50.15999984741211,
#         'lon': -96.07000732421875,
#         'alt': 0.0,
#         'cadence': 3,
#         'resolution': (256, 256),
#     }
#     return


##########################################
############# TEST EXAMPLES ##############
##########################################
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_rego_fisheye_map_example'], tol=20, remove_text=True, extensions=['png']
)
def test_rego_fisheye_map_example():
    """
    Plot a fisheye lens image and map it onto a map.
    """
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi
    import asilib.map

    location_code = 'RANK'
    time = datetime(2017, 9, 15, 2, 34, 0)
    alt_km = 110
    fig = plt.figure(figsize=(10, 4))
    ax = fig.add_subplot(121)
    bx = asilib.map.create_simple_map(
        fig_ax=(fig, 122), lon_bounds=(-102, -82), lat_bounds=(58, 68)
    )
    asi = asilib.asi.rego(location_code, time=time, alt=alt_km)
    asi.plot_fisheye(ax=ax)
    asi.plot_map(ax=bx)
    plt.tight_layout()
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_rego_keogram_example'], tol=20, remove_text=True, extensions=['png']
)
def test_rego_keogram_example():
    """
    Plot a keogram.
    """
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi

    asi = asilib.asi.rego(
        'RANK', time_range=(datetime(2017, 9, 15, 2, 30, 0), datetime(2017, 9, 15, 2, 35, 0))
    )
    ax, im = asi.plot_keogram()
    plt.colorbar(im)
    ax.axis('off')
    return



from typing import Tuple

import numpy as np

import pymap3d


Re = 6378.14  # km

def geodetic_skymap(
        imager_lla:Tuple[float],
        az_skymap:np.ndarray, 
        el_skymap:np.ndarray, 
        alt:float
        ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Project each pixel's (az, el) coordinates to geodetic (lat, lon) coordinates at an emission
    altitude.

    Parameters
    ----------
    imager_lla:tuple
        The imager's latitude, longitude, and altitude in km.
    az_skymap:np.ndarray 
        A 2d array of pixel azimuths. 
    el_skymap:np.ndarray 
        A 2d array of pixel elevations.
    alt:float
        Auroral emission altitude in km to map to.

    Returns
    -------
    np.ndarray
        The latitude skymap.
    np.ndarray
        The longitude skymap.
    """
    lat_skymap = np.zeros_like(az_skymap)
    lon_skymap = np.zeros_like(az_skymap)
    _alts = np.zeros_like(az_skymap)

    el_skymap[el_skymap <= 0] = np.nan

    # Range from observer to target assuming Earth is spherical, i.e. not an ellipsoid.
    # https://github.com/space-physics/dascasi/blob/4d72aa91e471a495566044c3fc387344dd12461f/src/dascasi/io.py#L107C32-L107C32
    _el_rad = np.deg2rad(el_skymap)
    _range = np.sqrt(
        (Re+alt)**2 +
        (Re+imager_lla[-1])**2 - 
        2*(Re+alt)*(Re+imager_lla[-1])*np.sin(
            _el_rad + np.arcsin((Re+imager_lla[-1])/(Re+alt)*np.cos(_el_rad))
            )
        )

    for row in range(az_skymap.shape[0]):
        for col in range(az_skymap.shape[1]):
            # Based on Michael Hirsh's (scivision) dascasi package.
            lat_skymap[row, col], lon_skymap[row, col], _alts[row, col] = pymap3d.aer2geodetic(
                az_skymap[row, col],
                el_skymap[row, col],
                _range[row, col] * 1e3,
                *imager_lla[:2], 
                1E3*imager_lla[-1],
                deg=True,
            )
    return lat_skymap, lon_skymap


from datetime import datetime

import numpy as np
import matplotlib.pyplot as plt
import asilib.map
from asilib.asi.psa_project import psa_project


n_avg = 10  # Average 10 image frames (the effective imager cadence is 10 Hz)

asi = psa_project(
    'vee',
    time_range=(datetime(2022, 3, 5, 11, 0), datetime(2022, 3, 5, 11, 2)),
    n_avg=n_avg
)

times, images = asi.data

fig, ax = plt.subplots(figsize=(5, 5), constrained_layout=True)
asi.animate_fisheye(ax=ax, ffmpeg_params={'framerate':100})


"""
Plot geographic maps using cartopy or the simple built-in function. Before you
project an ASI image onto a map, you will need to create a map using the 
following functions.

The simplest way to create a map is via :py:meth:`~asilib.map.create_map` that
by default creates a map above North America. :py:meth:`~asilib.map.create_map` 
is a wrapper that automatically chooses what library to plot the map: cartopy 
if it is installed, or asilib's :py:meth:`~asilib.map.create_simple_map` 
function to create a simple map if cartopy is not installed. All of these functions
output the subplot object with the map.

You can override this automatic behavior by calling the underlying functions directly:
:py:meth:`~asilib.map.create_simple_map` or :py:meth:`~asilib.map.create_cartopy_map`.

The two functions are called similarly with the exception of the ``ax``  kwarg if you 
need to specify what subplot to create the map on:

- for :py:meth:`~asilib.map.create_simple_map` you must pass in a ``ax`` subplot object, and 
- for :py:meth:`~asilib.map.create_cartopy_map`, you need to pass in a ``ax`` tuple containing two elements. The first element is the ```plt.Figure`` object, and the second element is a 3 digit number (or a tuple) specifying where to place that subplot. For example,

    .. code-block:: python

        >>> fig = plt.Figure()
        >>> ax = asilib.map.create_cartopy_map(ax=(fig, 111))
"""

import pathlib
from typing import List, Union, Tuple
import zipfile

import shapefile
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec
import numpy as np
import numpy.ma

try:
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature

    cartopy_imported = True
except ImportError as err:
    # You can also get a ModuleNotFoundError if cartopy is not installed
    # (as compared to failed to import), but it is a subclass of ImportError.
    cartopy_imported = False

import asilib


def create_map(
    lon_bounds: Tuple = (-160, -50),
    lat_bounds: Tuple = (40, 82),
    fig_ax: Tuple[matplotlib.figure.Figure, int] = None,
    coast_color: str = 'k',
    land_color: str = 'g',
    ocean_color: str = 'w',
):
    """
    Create a geographic map using cartopy (if installed) or asilib's own
    map library.

    Parameters
    ----------
    lon_bounds: tuple
        The map's longitude bounds.
    lat_bounds: tuple
        The map's latitude bounds.
    fig_ax: Tuple[matplotlib.figure.Figure, int]
        A two element tuple. First element is a ``matplotlib.figure.Figure`` object and second element is the
        subplot index or matplotlib.gridspec.SubplotSpec object. The second element is passed directly as ``args``
        `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.

        For example:

        .. code-block:: python

            fig = plt.Figure()
            ax = asilib.map.create_cartopy_map(fig_ax=(fig, 111))
    coast_color: str
        The coast color. If None will not draw it.
    land_color: str
        The land color. If None will not draw it.
    ocean_color: str
        The ocean color. If None will not draw it.

    Returns
    -------
    plt.Axes
        The subplot object containing the map.

    Examples
    --------
    These examples will differ if you have cartopy installed or not. To force a consistent map,
    replace ``asilib.map.create_map`` with ``asilib.map.create_simple_map`` or
    ``asilib.map.create_cartopy_map``.

    Examples
    --------
    >>> # Create a map above Scandinavia in a single subplot
    >>> import asilib.map
    >>> import matplotlib.pyplot as plt
    >>> ax = asilib.map.create_map(lon_bounds=[0, 38], lat_bounds=[50, 75])
    >>> ax.set_title('Generated via asilib.map.create_map()')
    >>> plt.show()

    >>> # The above examples made a map on one subplot. But what if you have multiple subplots?
    >>> import asilib.map
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure(figsize=(6, 10))
    >>> bx = asilib.map.create_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, 211))
    >>> cx = fig.add_subplot(2, 1, 2)
    >>> cx.plot(np.arange(10), np.arange(10))
    >>> fig.suptitle('Two subplots with equal sizes')
    >>> plt.show()

    >>> # Another multi-subplot example with different height ratios. The syntax is the same as in plt.subplot() (See the args section in https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplot.html).
    >>> import asilib.map
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure(figsize=(6, 10))
    >>> dx = (3, 1, (1, 2))
    >>> dx = asilib.map.create_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, dx))
    >>> ex = fig.add_subplot(3, 1, 3)
    >>> ex.plot(np.arange(10), np.arange(10))
    >>> fig.suptitle('Two subplots with unequal sizes')
    >>> plt.show()

    >>> # And make a map using gridspec
    >>> import asilib.map
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib.gridspec
    >>> fig = plt.figure()
    >>> gs = matplotlib.gridspec.GridSpec(1, 1, fig)
    >>> dx = asilib.map.create_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, gs))
    >>> dx.set_title('Map made using gridspec')
    >>> plt.show()

    See Also
    --------
    :py:meth:`~asilib.map.create_simple_map`:
        Create a simple map using asilib's own map library.
    :py:meth:`~asilib.map.create_cartopy_map`:
        Create a map using the cartopy library.
    """
    if cartopy_imported:
        ax = create_cartopy_map(
            lon_bounds=lon_bounds,
            lat_bounds=lat_bounds,
            fig_ax=fig_ax,
            coast_color=coast_color,
            land_color=land_color,
            ocean_color=ocean_color,
        )
    else:
        ax = create_simple_map(
            lon_bounds=lon_bounds,
            lat_bounds=lat_bounds,
            fig_ax=fig_ax,
            coast_color=coast_color,
            land_color=land_color,
            ocean_color=ocean_color,
        )
    if fig_ax is None:
        plt.tight_layout()
    return ax


def create_simple_map(
    lon_bounds: tuple = (-140, -60),
    lat_bounds: tuple = (40, 82),
    fig_ax: Tuple[matplotlib.figure.Figure, int] = None,
    coast_color: str = 'k',
    land_color: str = 'g',
    ocean_color: str = 'w',
    file: Union[str, pathlib.Path] = 'ne_10m_land',
) -> plt.Axes:
    """
    Create a simple map without cartopy.

    Parameters
    ----------
    lon_bounds: tuple
        The map's longitude bounds.
    lat_bounds: tuple
        The map's latitude bounds.
    fig_ax: Tuple[matplotlib.figure.Figure, int]
        A two element tuple. First element is a ``matplotlib.figure.Figure`` object and second element is the
        subplot index or matplotlib.gridspec.SubplotSpec object. The second element is passed directly as ``args``
        `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.

        For example:

        .. code-block:: python

            fig = plt.Figure()
            ax = asilib.map.create_simple_map(fig_ax=(fig, 111))
    coast_color: str
        The coast color. If None will not draw it.
    land_color: str
        The land color. If None will not draw it.
    ocean_color: str
        The ocean color. If None will not draw it.
    file: str or pathlib.Path
        The path to the shapefile zip archive. If str, it will try to load the
        shapefile in asilib/data/{file}. You can download other shapefiles from
        https://www.naturalearthdata.com/downloads/10m-physical-vectors/.

    Examples
    --------
    See :py:meth:`~asilib.map.create_map` and replace ``create_map`` with ``create_simple_map``.

    Returns
    -------
    plt.Axes
        The subplot object containing the map.
    """
    _validate_fig_ax(fig_ax)
    shp_path = asilib.config['ASILIB_DIR'] / 'data' / f'{file}'

    with zipfile.ZipFile(shp_path.with_suffix('.zip'), 'r') as archive:
        shp = archive.open(f'{file}.shp', "r")
        dbf = archive.open(f'{file}.dbf', "r")
        sf = shapefile.Reader(shp=shp, dbf=dbf)
        i = 0  # I'm unsure what the other shapes are, but i=0 works.
        lats = np.array([point[0] for point in sf.shapes()[i].points])
        lons = np.array([point[1] for point in sf.shapes()[i].points])

    # Since the landmass shapes are represented continuously in (lats, lons),
    # matplotlib draws straight (annoying) lines between them. This code uses
    # the jumps bool array and masked_arrays to remove those lines.
    jumps = (np.abs(lats[1:] - lats[:-1]) > 3) | (np.abs(lons[1:] - lons[:-1]) > 3)
    mlats = numpy.ma.masked_array(lats[:-1], mask=jumps)
    mlons = numpy.ma.masked_array(lons[:-1], mask=jumps)

    split_lats = _consecutive(lats, jumps)
    split_lons = _consecutive(lons, jumps)

    if fig_ax is None:
        _, ax = plt.subplots()
    else:
        fig = fig_ax[0]
        if isinstance(fig_ax[1], (int, matplotlib.gridspec.SubplotSpec)):
            ax = fig.add_subplot(fig_ax[1])
        else:
            ax = fig.add_subplot(*fig_ax[1])

    if ocean_color is not None:
        ax.set_facecolor(ocean_color)
    if coast_color is not None:
        ax.plot(mlats, mlons, coast_color)
    if land_color is not None:
        for split_lon, split_lat in zip(split_lons, split_lats):
            ax.fill(split_lat, split_lon, land_color, zorder=0)

    ax.set_xlim(lon_bounds)
    ax.set_ylim(lat_bounds)
    return ax


def create_cartopy_map(
    lon_bounds: tuple = (-160, -50),
    lat_bounds: tuple = (40, 82),
    fig_ax: Tuple[matplotlib.figure.Figure, int] = None,
    coast_color: str = 'k',
    land_color: str = 'g',
    ocean_color: str = 'w',
) -> plt.Axes:
    """
    A helper function to create two map styles: a simple black and white map, and
    a more sophisticated map with green land.

    Parameters
    ----------
    lon_bounds: tuple or list
        A tuple of length 2 specifying the map's longitude bounds.
    lat_bounds: tuple or list
        A tuple of length 2 specifying the map's latitude bounds.
    fig_ax: Tuple[matplotlib.figure.Figure, int]
        A two element tuple. First element is a ``matplotlib.figure.Figure`` object and second element is the
        subplot index or matplotlib.gridspec.SubplotSpec object. The second element is passed directly as ``args``
        `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.

        For example:

        .. code-block:: python

            fig = plt.Figure()
            ax = asilib.map.create_cartopy_map(fig_ax=(fig, 111))

    coast_color: str
        The coast color. If None will not draw it.
    land_color: str
        The land color. If None will not draw it.
    ocean_color: str
        The ocean color. If None will not draw it.

    Returns
    -------
    ax: plt.Axes
        The subplot object with a cartopy map.

    Examples
    --------
    See :py:meth:`~asilib.map.create_map` and replace ``create_map`` with ``create_simple_map``.
    """
    _validate_fig_ax(fig_ax)
    plot_extent = [*lon_bounds, *lat_bounds]
    central_lon = np.mean(lon_bounds)
    central_lat = np.mean(lat_bounds)
    projection = ccrs.Orthographic(central_lon, central_lat)

    if fig_ax is None:
        fig = plt.figure(figsize=(8, 5))
        _ax = fig.add_subplot(111, projection=projection)
    else:
        fig = fig_ax[0]
        if isinstance(fig_ax[1], (int, matplotlib.gridspec.SubplotSpec)):
            _ax = fig.add_subplot(fig_ax[1], projection=projection)
        else:
            _ax = fig.add_subplot(*fig_ax[1], projection=projection)
    if land_color is not None:
        _ax.add_feature(cfeature.LAND, color=land_color)
    if ocean_color is not None:
        _ax.add_feature(cfeature.OCEAN, color=ocean_color)
    if coast_color is not None:
        _ax.add_feature(cfeature.COASTLINE, edgecolor=coast_color)
    _ax.gridlines(linestyle=':')
    _ax.set_extent(plot_extent, crs=ccrs.PlateCarree())
    return _ax


def _validate_fig_ax(fig_ax):
    if fig_ax is not None:
        valid_first_param = isinstance(fig_ax[0], matplotlib.figure.Figure)
        valid_second_param = isinstance(
            fig_ax[1], (int, matplotlib.gridspec.GridSpec, matplotlib.gridspec.SubplotSpec, tuple)
        )
        if valid_first_param and valid_second_param:
            return
        elif valid_first_param and not valid_second_param:
            raise ValueError('Second element of fig_ax must be a integer or a plt.gridspec object.')
        elif valid_second_param and not valid_first_param:
            raise ValueError('First element of fig_ax must be a plt.figure object.')
        else:
            raise ValueError(
                f'fig_ax must be a 2-element tuple with plt.figure as the first element'
                f'and either an integer or plt.gridspec object as the second element.'
            )
    return


def _consecutive(data, jump_bool):
    """
    Calculate where the array jumps.

    Taken from: https://stackoverflow.com/questions/7352684/
    how-to-find-the-groups-of-consecutive-elements-in-a-numpy-array
    """
    return np.split(data, np.where(jump_bool)[0] + 1)


if __name__ == '__main__':
    import asilib.map
    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.gridspec

    # Create a simple map above Scandinavia in a single plot
    ax = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75])
    ax.set_title('Generated via asilib.map.create_simple_map()')

    # The above examples made a map on one subplot. But what if you have multiple subplots?
    fig = plt.figure(figsize=(6, 10))
    bx = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, 211))
    cx = fig.add_subplot(2, 1, 2)
    cx.plot(np.arange(10), np.random.rand(10))
    fig.suptitle('Two subplots with equal aspect ratios')

    # Another multi-subplot example with different height ratios. The syntax is the same as
    # in plt.subplot() (See the args section in
    # https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplot.html).
    fig = plt.figure(figsize=(6, 10))
    dx = (3, 1, (1, 2))
    dx = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, dx))
    ex = fig.add_subplot(3, 1, 3)
    ex.plot(np.arange(10), np.random.rand(10))
    fig.suptitle('Two subplots with differing aspect ratios')

    # And make a map using gridspec
    fig = plt.figure()
    gs = matplotlib.gridspec.GridSpec(1, 1, fig)
    dx = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, gs))
    dx.set_title('Map made using gridspec')
    plt.show()



"""
Test the mango loaders and a few basic loading functions.
"""

from datetime import datetime

import requests
import pytest
import matplotlib.testing.decorators

from asilib.asi import mango

def test_mango_redline_time():
    """
    Get one timestamp from the redline imager.
    """
    time=datetime(2024, 9, 24, 3, 56, 0)
    location_code='CVO'
    asi = mango(location_code, 'redline', time=time)
    assert asi.data.time == time
    assert asi.data.image.shape == (500, 500)
    return

def test_mango_greenline_time():
    """
    Get one timestamp from the greenline imager.
    """
    time=datetime(2024, 10, 25, 4, 20, 0)
    location_code='BDR'
    asi = mango(location_code, 'greenline', time=time)
    assert asi.data.time == time
    assert asi.data.image.shape == (500, 500)
    return

def test_mango_redline_time_range():
    """
    Get multiple timestamps from the redline imager.
    """
    time_range=(datetime(2021, 11, 4, 1, 0), datetime(2021, 11, 4, 12, 24))
    location_code='CFS'
    asi = mango(location_code, 'redline', time_range=time_range)
    assert asi.data.time.shape[0] == 158
    assert asi.data.time[0] == datetime(2021, 11, 4, 1, 56)
    assert asi.data.time[-1] == time_range[-1]
    return

def test_mango_greenline_time_range():
    """
    Get multiple timestamps from the greenline imager.
    """
    time_range=(datetime(2024, 10, 25, 1, 50, 0), datetime(2024, 10, 25, 11, 22, 0))
    location_code='BDR'
    asi = mango(location_code, 'greenline', time_range=time_range)
    assert asi.data.time[0] == time_range[0]
    assert asi.data.time[-1] == time_range[-1]
    assert asi.data.image.shape == (287, 500, 500)
    return


"""
An ASI for testing asilib.Imager.
"""
import pandas as pd
import numpy as np
from datetime import datetime, timedelta

import matplotlib.pyplot as plt

import asilib
import asilib.utils as utils


def fake_asi(
    location_code: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    alt: int = 110,
    pixel_center: bool = True,
) -> asilib.Imager:
    """
    Create an Imager instance with the fake_asi images and skymaps.

    Parameters
    ----------
    location_code: str
        The ASI's location code (four letters). Can be either GILL, ATHA, or TPAS.
        Case insensitive.
    time: str or datetime.datetime
        A time to look for the ASI data at. Either the time or the time_range
        must be specified.
    time_range: list of str or datetime.datetime
        A length 2 list of string-formatted times or datetimes to bracket
        the ASI data time interval.
    alt: int
        The reference skymap altitude, in kilometers.
    pixel_center: bool
        If True, then the skymap specifies the pixel centers, otherwise the skymap specifies
        the pixel vertices. Specifying the vertices more accurately describes how the pixels
        field of view transforms from a square to a polynomial.

    Returns
    -------
    :py:meth:`~asilib.imager.Imager`
        The an Imager instance with the fake_asi data.
    """
    location_code = location_code.upper()
    locations = asi_info()
    _location = locations.loc[locations.index == location_code, :]

    meta = get_meta(_location)
    skymap = get_skymap(meta, alt=alt, pixel_center=pixel_center)
    file_info = get_file_info(meta, time=time, time_range=time_range)
    plot_settings = {'color_bounds':(1, 255)}
    return asilib.Imager(file_info, meta, skymap, plot_settings=plot_settings)


def asi_info() -> pd.DataFrame:
    """
    The test ASI has three locations, a, b, and c.
    """
    locations = pd.DataFrame(
        data=np.array(
            [
                [56.3494, -94.7056, 500],
                [54.7213, -113.285, 1000],
                [53.8255, -101.2427, 0],
            ]
        ),
        columns=['lat', 'lon', 'alt'],
        index=['GILL', 'ATHA', 'TPAS'],
    )
    return locations


def get_meta(location_dict):
    """
    Get the ASI metadata.
    """
    meta = {
        'array': 'TEST',
        'location': location_dict.index[0],
        'lat': location_dict['lat'].iat[0],
        'lon': location_dict['lon'].iat[0],
        'alt': location_dict['alt'].iat[0] / 1e3,  # km
        'cadence': 10,
        'resolution': (512, 512),
    }
    return meta


def get_skymap(meta: dict, alt: int, pixel_center: bool = True):
    """
    Create a skymap based on the ASI location in the metadata.

    Parameters
    ----------
    meta: dict
        The ASI metadata with the imager resolution and cadence.
    alt: int
        The reference skymap altitude, in kilometers.
    pixel_center: bool
        If True, then the skymap specifies the pixel centers, otherwise the skymap specifies
        the pixel vertices. Specifying the vertices more accurately describes how the pixels
        field of view transforms from a square to a polynomial.
    """
    assert alt in [90, 110, 150], (
        f'The {alt} km altitude does not have a corresponding map: '
        'valid_altitudes=[90, 110, 150].'
    )
    skymap = {}
    lon_bounds = [meta['lon'] - 10 * (alt / 110), meta['lon'] + 10 * (alt / 110)]
    lat_bounds = [meta['lat'] - 10 * (alt / 110), meta['lat'] + 10 * (alt / 110)]

    if pixel_center:
        pad = 0
    else:
        pad = 1

    # TODO: Add tests for skymaps specifying pixel edges too.
    # lon_bounds[::-1] so that east is to the right.
    _lons, _lats = np.meshgrid(
        np.linspace(*lon_bounds[::-1], num=meta['resolution'][0] + pad),
        np.linspace(*lat_bounds, num=meta['resolution'][1] + pad),
    )
    std = 5 * (alt / 110)
    dst = np.sqrt((_lons - meta['lon']) ** 2 + (_lats - meta['lat']) ** 2)
    # the 105 multiplier could be 90, but I chose 105 (and -15 offset) to make the
    # skymap realistic: the edges are NaNs.
    elevations = 105 * np.exp(-(dst**2) / (2.0 * std**2)) - 15
    elevations[elevations < 0] = np.nan
    # These are the required skymap keys for asilib.Imager to work.
    skymap['el'] = elevations
    skymap['alt'] = alt
    skymap['lon'] = _lons
    skymap['lon'][~np.isfinite(skymap['el'])] = np.nan
    skymap['lat'] = _lats
    skymap['lat'][~np.isfinite(skymap['el'])] = np.nan
    skymap['path'] = __file__

    # Calculate the azimuthal angle using cross product between a northward-pointing unit vector
    # and the (_lons, _lats) grid. See https://stackoverflow.com/a/16544330 for an explanation.
    dot_product = 0 * (_lons - meta['lon']) + 1 * (_lats - meta['lat'])
    determinant = 0 * (_lats - meta['lat']) - 1 * (_lons - meta['lon'])
    skymap['az'] = (180 / np.pi) * np.arctan2(determinant, dot_product)
    # transform so it goes 0->360 in a clockwise direction.
    skymap['az'] = -1 * skymap['az']
    skymap['az'][skymap['az'] < 0] = 360 + skymap['az'][skymap['az'] < 0]
    return skymap


def get_file_info(
    meta: dict, time: utils._time_type = None, time_range: utils._time_range_type = None
) -> dict:
    """
    Get some images and time stamps. One image and time stamp if time is specified,
    or multiple images if time_range is specified.

    Parameters
    ----------
    meta: dict
        The ASI metadata with the imager resolution and cadence.
    time: str or datetime.datetime
        A time to look for the ASI data at. Either the time or the time_range
        must be specified.
    time_range: list of str or datetime.datetime
        A length 2 list of string-formatted times or datetimes to bracket
        the ASI data time interval.
    """
    if (time is not None) and (time_range is not None):
        raise ValueError("time and time_range can't be simultaneously specified.")
    if (time is None) and (time_range is None):
        raise ValueError("Either time or time_range must be specified.")

    if time is not None:
        time = utils.validate_time(time)
        file_path = _get_file_path(meta, time)
        image_times, images = _data_loader(file_path)
        start_file_time = time.replace(minute=0, second=0, microsecond=0)

        _data = {
            'time': time,
            'path': [file_path],
            'start_time': [start_file_time],
            'end_time': [start_file_time + timedelta(hours=1)],
            'loader': _data_loader,
        }

    else:
        time_range = utils.validate_time_range(time_range)
        start_file_time = time_range[0].replace(minute=0, second=0, microsecond=0)
        hours = (
            int((time_range[1] - time_range[0]).total_seconds() // 3600) + 1
        )  # +1 to load the final hour.

        # These are all of the keys required by asilib.Imager.
        _data = {
            'time_range': time_range,
            'path': [
                _get_file_path(meta, start_file_time + timedelta(hours=i)) for i in range(hours)
            ],
            'start_time': [start_file_time + timedelta(hours=i) for i in range(hours)],
            'end_time': [start_file_time + timedelta(hours=1 + i) for i in range(hours)],
            'loader': _data_loader,
        }
    return _data


def _data_loader(file_path):
    """
    Given a file_path, open the file and return the time stamps and images.

    Time stamps are every 10 seconds. The images are all 0s except a horizontal and a vertical
    line whose position is determined by seconds_since_start modulo resolution (512).
    """
    # Assume that the image time stamps are at exact seconds
    date_time_str = '_'.join(file_path.split('_')[:2])
    file_time = datetime.strptime(date_time_str, '%Y%m%d_%H%M%S')

    times = np.array(
        [file_time + timedelta(seconds=i * 10) for i in range(3600 // 10)]
    )  # 10 s cadence

    images = np.ones((times.shape[0], 512, 512))
    for i, time in enumerate(times):
        sec = (time - file_time).total_seconds()
        images[i, int(sec % 512) - 10 : int(sec % 512) + 10, :] = 255
        images[i, :, int(sec % 512) - 10 : int(sec % 512) + 10] = 100
    return times, images


def _get_file_path(meta, time):
    return f'{time:%Y%m%d_%H}0000_{meta["location"]}_fake_asi.png'  # does not exist.


def plot_skymap(location_code, alt=110, pixel_center=True):
    """
    Visualize the skymap to get a better idea on what a realistic one looks like
    (if perfectly aligned).
    """
    location_code = location_code.upper()
    locations = asi_info()
    _location = locations.loc[locations.index == location_code, :]
    meta = get_meta(_location)

    skymap = get_skymap(meta, alt=alt, pixel_center=pixel_center)
    keys = ['el', 'az', 'lat', 'lon']
    fig, ax = plt.subplots(1, len(keys), sharex=True, sharey=True, figsize=(3.7 * len(keys), 4))

    for ax_i, key in zip(ax, keys):
        p = ax_i.pcolormesh(skymap[key])
        plt.colorbar(p, ax=ax_i)
        ax_i.set_title(key)
    plt.suptitle('asilib | test ASI skymap')
    plt.tight_layout()
    return


if __name__ == '__main__':
    # asi = fake_asi('GILL', time_range=('2015-01-01T15:00:15.17', '2015-01-01T20:00'))
    # asi = fake_asi('GILL', time='2015-01-01T15:14:00.17')
    # asi.plot_fisheye(color_bounds=(1, 255), origin=(0.85, 0.15), cardinal_directions='NEWS')
    # plt.show()
    asi = fake_asi('GILL', time='2015-01-01T15:14:00.17')
